(*********************************************************************

  HOC04DefLTS

**********************************************************************

  Copyright (C) 2014
  Simon Boulier, Martin Escarra, Petar Maksimovic, Alan Schmitt

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************)

Require Import HOC01Defs.
Require Import HOC02BaseLemmas.
Require Import HOC03CanonicalLemmas.

(**********************************
  TRANSITION LABELS
 **********************************)

Inductive label : Type :=
 | LabOut : chan -> wfprocess -> label
 | LabIn  : chan -> label
 | LabRem : var  -> label
 | Tau.

(**********************************
  ABSTRACTIONS AND AGENTS
 **********************************)

Inductive abstraction : Set :=
 | AbsPure : var -> wfprocess -> abstraction
 | AbsPar1 : abstraction -> wfprocess -> abstraction
 | AbsPar2 : wfprocess -> abstraction -> abstraction.

(* Agents are either abstractions or processes *)
Inductive agent : Set :=
 | AA : abstraction -> agent
 | AP : wfprocess -> agent.

(* Left parallelization of agents *)
Definition par_ag1 (a : agent) (p : wfprocess) : agent :=
  match a with
    | AA f => AA (AbsPar1 f p)
    | AP q => AP (wfp_Par q p)
  end.

(* Right parallelization of agents *)
Definition par_ag2 (p : wfprocess) (a : agent) : agent :=
  match a with
    | AA f => AA (AbsPar2 p f)
    | AP q => AP (wfp_Par p q)
  end.

(* Constant Abstractions *)
Inductive const_abs : abstraction -> Prop :=
 | CAbs     : forall X (p : wfprocess), X \notin GV p -> const_abs (AbsPure X p)
 | CAbsPar1 : forall f p, const_abs f -> const_abs (AbsPar1 f p)
 | CAbsPar2 : forall f p, const_abs f -> const_abs (AbsPar2 p f).

Hint Resolve CAbs CAbsPar1 CAbsPar2.

(* Decision procedure for abstraction constance *)
Fixpoint const_abs_dec (fp : abstraction): bool :=
  match fp with
    | AbsPure X p => ifb (X # p) then true else false                 
    | AbsPar1 fp p  => const_abs_dec fp
    | AbsPar2 p  fp => const_abs_dec fp
  end.

(* Finding out if an abstraction is constant is decidable *)
Instance const_abs_decidable : forall (f:abstraction), Decidable (const_abs f).  
Proof.
  intros.  
  applys~ decidable_make (const_abs_dec f).
  inductions f; simpls~. 
  cases_if; try rename f into Hf; symmetry.
    applys~ isTrue_true. 
    apply isTrue_false. intro H1. inverts~ H1. 
    rewrite isTrue_def in *. cases_if; cases_if~.
      inverts H. false.
      false~ H. 
    rewrite isTrue_def in *. cases_if; cases_if~.
      inverts H. false.
      false~ H. 
Qed.

(* Size of abstractions *)
Fixpoint sizeAbs (fp :abstraction) : nat :=
  match fp with
    | AbsPure X p => sizeX X p                     
    | AbsPar1 fp p  => sizeAbs fp
    | AbsPar2 p  fp => sizeAbs fp
  end.

(* Instantiation of abstractions *)
Fixpoint wfp_inst_Abs (fp : abstraction) (q : wfprocess) : wfprocess :=
  match fp with
    | AbsPure X p => wfp_subst q X p                  
    | AbsPar1 fp p => wfp_Par (wfp_inst_Abs fp q) p
    | AbsPar2 p  fp => wfp_Par p (wfp_inst_Abs fp q)
  end. 

(**********************************
  THE TRANSITION SYSTEM
 **********************************)

Reserved Notation "{{ p -- l ->> ap }}" (at level 60).

Inductive transition : wfprocess -> label -> agent -> Prop :=
 | TrOut : forall a p, {{wfp_Send a p -- LabOut a p ->> AP wfp_Nil}}
 | TrIn  : forall a X p, {{wfp_Abs a X p -- LabIn a ->> AA (AbsPure X p)}}
 | TrRem : forall X, {{wfp_Gvar X -- LabRem X ->> AP wfp_Nil}}
 | TrAct1 : forall p q l ap, (transition p l ap) -> (transition (wfp_Par p q) l (par_ag1 ap q))
 | TrAct2 : forall p q l aq, (transition q l aq) -> (transition (wfp_Par p q) l (par_ag2 p aq))
 | TrTau1 : forall a p p' q p'' fq, 
    (transition p (LabOut a p'') (AP p')) -> 
    (transition q (LabIn a)    (AA fq)) -> 
    (transition (wfp_Par p q) Tau (AP (wfp_Par p' (wfp_inst_Abs fq p''))))
 | TrTau2 : forall a p q q' q'' fp, 
    (transition p (LabIn a) (AA fp)) -> 
    (transition q (LabOut a q'') (AP q')) -> 
    (transition (wfp_Par p q) Tau (AP (wfp_Par (wfp_inst_Abs fp q'') q')))
where "{{ p -- l ->> ap }}" := (transition p l ap).

Hint Resolve TrOut TrIn TrRem TrAct1 TrAct2 TrTau1 TrTau2.



